require 'class'

--MAManager is a 'manager' class for MapActions
MAManager = class(function(se)
	se.allActions = {} --List of all map actions
	se.activeActions = {} --List of active map actions
	se.actionUpdatesPerTick = 10 --Number of active actions to update per tick
	se.lastUpdatedActionID = nil --ID number of last updated action
	se.nextID = 1 --Unique IDs for instantiated actions
end)

--For now, just activate "initially active" map actions
function MAManager:initializeActions()
	for k, action in pairs(self.allActions) do
		if action.initiallyActive then
			action:activate()
		end
	end
end

function MAManager:update(deltaTime)
	local actionsUpdated = 0
	--look through our activeActions, starting at the lastUpdatedActionID, and update them
	for id, action in next, self.activeActions, self.lastUpdatedActionID do
		action:update(deltaTime)
		self.lastUpdatedActionID = id
		actionsUpdated = actionsUpdated + 1
		if actionsUpdated >= self.actionUpdatesPerTick then break end
	end
	--if we ran out of actions without updating the maximum number, start over at the beginning of the list
	if actionsUpdated < self.actionUpdatesPerTick then
		self.lastUpdatedActionID = nil
		for id, action in next, self.activeActions, self.lastUpdatedActionID do
			action:update(deltaTime)
			self.lastUpdatedActionID = id
			actionsUpdated = actionsUpdated + 1
			if actionsUpdated >= self.actionUpdatesPerTick then return end
		end
	end
end

--Create our singleton ScriptEngine
ScriptEngine = MAManager()




--MapAction Enumerations
MAExpirationMode = {
	onTimer = 0, --onTimer means the action will only expire at the end of its timer
	onExecution = 1, --onExecution means the action was executed (i.e. a TUNI unit search was done).
	onSuccessfulExecution = 2, --onSuccessfulExecution means the action reports it executed successfully.
	onFailedExecution = 3, --onFailedExecution means the action reports it did not execute sucessfully.
	never = 4 --this action will never expire. to kill it, it must be deactivated by another action.
}
MAResult = {
	neverExecuted = 0,
	succeeded = 1,
	failed = 2,
	raisedError = 3
}

--MapAction base class
MapAction = class(function(ma)
	--USER ACCESSIBLE DATA
	ma.expirationMode = MAExpirationMode.onExecution
	ma.timerStart = 0
	ma.timerEnd = 0
	ma.timerRandomDelay = 0
	ma.activatesOnlyOnce = false
	ma.initiallyActive = false
	ma.noInitialDelay = false
	ma.onlyInitialDelay = false
	ma.deletedOnDeactivation = false
	
	--THESE TABLES CAN BE CONDITIONALLY ADDED BY THE USER
	--[[ tables of actions to conditionally activate
	ma.activatesOnActivation = {}
	ma.activatesOnDeactivation = {}
	ma.activatesOnExecution = {}
	ma.activatesOnSuccess = {}
	ma.activatesOnFailure = {}
	ma.activatesOnError = {}
	ma.activatesOnInhibited = {}
	ma.activatesOnMissingPrerequisites = {} ]]--
	
	--[[ tables of actions to conditionally deactivate
	ma.deactivatesOnActivation = {}
	ma.deactivatesOnDeactivation = {}
	ma.deactivatesOnExecution = {}
	ma.deactivatesOnSuccess = {}
	ma.deactivatesOnFailure = {}
	ma.deactivatesOnError = {} ]]--
	
	--[[tables of misc other actions
	ma.inhibitions = {}
	ma.prerequisites = {} ]]--
	
	--USERS SHOULDN'T TOUCH THESE FIELDS
	ma.ID = ScriptEngine.nextID
	ma.active = false
	ma.secondsSinceActivation = 0
	ma.randomDelay = 0
	ma.result = MAResult.neverExecuted
	--adjust class static variables
	ScriptEngine.allActions[ma.ID] = ma
	ScriptEngine.nextID = ma.ID + 1
end)

function MapAction:activate()
	--don't activate if we're set to activate only once and already activated
	if not(activatesOnlyOnce and self.result ~= MAResult.neverTriggered) then
		--ensure we don't activate an active action
		if not self.active then
			self.active = true
			self.randomDelay = self.timerRandomDelay * math.random()
			self.secondsSinceActivation = 0
			ScriptEngine.activeActions[self.ID] = self
			--activate other actions
			if self.activatesOnActivation then self:activateActions(self.activatesOnActivation) end
			if self.deactivatesOnActivation then self:deactivateActions(self.deactivatesOnActivation) end
		end
	end
end

function MapAction:deactivate()
	--ensure we don't deactivate an inactive action
	if self.active then
		self.active = false
		ScriptEngine.activeActions[self.ID] = nil
		if self.activatesOnDeactivation then self:activateActions(self.activatesOnDeactivation) end
		if self.deactivatesOnDeactivation then self:deactivateActions(self.deactivatesOnDeactivation) end
	end
end

function MapAction:doExecute()
	return MAResult.succeeded --plain old MapActions do nothing but succeed. it's a good life.
end

function MapAction:checkTimer()
	--if we're not past the timer start delay, do nothing
	--we must check to see if we haven't executed yet and we want a start delay
	--this should account for "noInitialDelay" and "onlyInitialDelay" flags
	if self.result == MAResult.neverExecuted then
		if not self.noInitialDelay and self.secondsSinceActivation < self.timerStart + self.randomDelay then return false end
	elseif not self.onlyInitialDelay then
		if self.secondsSinceActivation < self.timerStart + self.randomDelay then return false end
	end
	
	--if we're on a timer and passed beyond timer end, deactivate self
	if self.expirationMode == MAExpirationMode.onTimer and self.secondsSinceActivation >= self.timerEnd + self.randomDelay then
		self:deactivate()
		return false
	--elseif we've got a valid timer end which we've passed, and our expiry mode isn't 'never', deactivate self
	elseif self.timerEnd ~= 0 and self.expirationMode ~= MAExpirationMode.never and self.secondsSinceActivation >= self.timerEnd + self.randomDelay then
		self:deactivate()
		return false
	end
	--passed all the tests
	return true
end

function MapAction:checkInhibitions()
	--check to see if this action has inhibitors
	if self.inhibitions and #(self.inhibitions) > 0 then
		--check if all are inhibitors are successful
		local inhibited = true
		for key, inhibitor in pairs(self.inhibitions) do
			if not inhibitor.result == MAResult.succeeded then
				inhibited = false
			end
		end
		--if all inhibitors are successful, do nothing
		if inhibited then
			if self.activatesOnInhibited then self:activateActions(self.activatesOnInhibited) end
			return false
		end
	end
	return true
end

function MapAction:checkPrerequisites()
	--check to see if this action has prerequisites
	if self.prerequisites and #(self.prerequisites) > 0 then
		--check if any prerequisite is successful
		local reqsMet = false
		for key, prerequisite in pairs(self.prerequisites) do
			if prerequisite.result == MAResult.succeeded then
				reqsMet = true
				break
			end
		end
		--if no prerequisite succeeded, do nothing
		if not reqsMet then
			if self.activatesOnMissingPrerequisites then self:activateActions(self.activatesOnMissingPrerequisites) end
			return false
		end
	end
	return true
end

function MapAction:activateActions(actions)
	for key, action in pairs(actions) do
		action:activate()
	end
end

function MapAction:deactivateActions(actions)
	for key, action in pairs(actions) do
		action:deactivate()
	end
end

function MapAction:update(deltaTime)
	--update the action's timer
	self.secondsSinceActivation = self.secondsSinceActivation + deltaTime
	--check if this action gets executed
	local shouldExecute = self:checkTimer() and self:checkInhibitions() and self:checkPrerequisites()
	if shouldExecute then
		--we've passed all the execution tests, so execute!
		self.result = self:doExecute()
		
		--act on result to (de)activate other MapActions
		if self.activatesOnExecution then self:activateActions(self.activatesOnExecution) end
		if self.deactivatesOnExecution then self:deactivateActions(self.deactivatesOnExecution) end
		--succeeded result
		if self.result == MAResult.succeeded then
			if self.activatesOnSuccess then self:activateActions(self.activatesOnSuccess) end
			if self.deactivatesOnSuccess then self:deactivateActions(self.deactivatesOnSuccess) end
		--failed result
		elseif self.result == MAResult.failed then
			if self.activatesOnFailure then self:activateActions(self.activatesOnFailure) end
			if self.deactivatesOnFailure then self:deactivateActions(self.deactivatesOnFailure) end
		--raisedError result
		elseif self.result == MAResult.raisedError then
			if self.activatesOnError then self:activateActions(self.activatesOnError) end
			if self.deactivatesOnError then self:deactivateActions(self.deactivatesOnError) end
		end
		
		--check for action expiration
		if self.expirationMode == MAExpirationMode.never then
			self:activate()
		elseif self.expirationMode == MAExpirationMode.onExecution then
			self:deactivate()
		elseif self.expirationMode == MAExpirationMode.onSuccessfulExecution and self.result == MAResult.succeeded then
			self:deactivate()
		elseif self.expirationMode == MAExpirationMode.onFailedExecution and self.result == MAResult.failed then
			self:deactivate()
		end
		
	end
	--and we're done updating the action
end


--TODO: need to recursively print tables, try using the example script found on the Lua website for printing the namespace
function MapAction:__tostring()
	local result = self:description()
	for key, value in pairs(self) do
		if type(key) == 'boolean' then if key then key = "true" else key = "false" end end
		if type(value) == 'boolean' then if value then value = "true" else value = "false" end end
		if type(value) == 'table' then value = 'table' end
		result = result .. "\n\t" .. key .. " = " .. value
	end
	result = result .. "\n"
	return result
end

function MapAction:description()
	return 'MapAction'
end